unit UnitIntGCent;

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                      UnitGCent par Kinzinger Ren                         //
//                        (version octobre 2009)                             //
//                                                                           //
//       Calculs sur des grands nombres entiers avec Delphi 32 bits          //
//       nombres cods en base 100                                           //
//                                                                           //
//       Code : EXPERIMENTAL  compatible Delphi5, Delphi7, Delphi2009        //
///////////////////////////////////////////////////////////////////////////////

interface

uses Dialogs, ClipBrd, Sysutils, uFun2;

type GCent = string;

var  nbToursGCent : longword;

// ROUTINES DE CONVERSIONS *****************************************************
function StrToGCent(St : string): GCent;
function GCentToStr(Nb : GCent) : string;
function IntToGCent(Nb: longword): GCent;
function GCentToInt64(Nb: GCent): Int64;

// ROUTINES OPERATOIRES DE BASE SUR LES STRING GCent ***************************
function  AddGCent(Nv1,Nv2: GCent): GCent;
function  Additionne(Nv1,Nv2: GCent): GCent;
function  DifGCent(Nv1,Nv2: GCent): GCent;
function  Soustrait(Nv1,Nv2: GCent): GCent;
function  Multiplie(Nv1,Nv2 : GCent): GCent;
function  Demi(Nv : GCent): GCent;
function  Double(Nv: GCent): GCent;
function  Incremente(Nv: GCent): GCent;
function  Decremente(Nv: GCent): GCent;
procedure DivModGCent(Nv1,Nv2 : GCent; var Q,R : GCent); // avec ShowMessage si Nv2 = Zro
procedure DivMod(Nv1,Nv2 : GCent; var Q,R : GCent);

// Calcule la racine carre entire par dfaut d'un GCent;
// Algorithme de Hron avec semence en potence
function RacineGCent(Nb : GCent): GCent;

// Calcule Nb^Exp et redonne un GCent
function  PuissanceGCent(Nb: GCent; Exp : longword): GCent;

// ROUTINES DE TESTS ***********************************************************
function  CompGCent(Nv1,Nv2 : GCent): shortInt;
function  GCentPair(Nv: GCent): boolean;

// ROUTINES DE GENERATIONS RAPIDES *********************************************
function  RandomGCent(NbDigits : longword; const Odd : boolean): GCent;

// ROUTINES des Oprations optimises sur GCent ********************************
function  GCentMulN10(Nv: GCent; N : longword): GCent;
function  GCentDivN10(Nv: GCent; N : longword): GCent;
function  NbZeroGCent(Nv: GCent): longword;

// renvoie un GCent multipli N fois par 100 (ultra rapide);
function  ShlNGCent(Nv : GCent; N : longword): Gcent; register
// renvoie un GCent divis N fois par 100 (ultra rapide);
function  ShrNGCent(Nv : GCent; N : longword): Gcent; register

procedure TrimD0GCent(var Nv : GCent; var nz : int64);

// Renvoie 10  la puissance Exp
function  DixPowerN(Exp : longword): GCent; register;

implementation

// ROUTINES INTERNES  **********************************************************

// corrige les 0 avant du buffer esi;eax
procedure SansZero;
asm
   @NoZero:
   cmp    eax,1
   je     @retour
   cmp    byte ptr[esi],0
   jne    @retour
   inc    esi
   dec    eax
   jmp    @NoZero
   @retour:
end;

// recopie trs vite esi dans edi sur longueur ecx
procedure Recopie;
asm
    push   edx
    mov    edx,ecx
    and    edx,$3
    shr    ecx,2
    cld
    rep    movsd
    mov    ecx,edx
    rep    movsb
    pop    edx
end;

// Compare deux buffers
// en entre esi;eax et edi;edx
// Retour par le registre des flags
// S:=1 => N1 < N2 ; Z:=1 => N1=N2; S et Z = 0 => N1 > N2
procedure compare;
asm
   push   eax
   push   edx
   mov    ecx,eax
   sub    eax,edx
   jnz    @fin
   xor    ecx,ecx
   @RCmp:
   mov    ah,byte ptr[esi]+ecx
   mov    al,byte ptr[edi]+ecx
   sub    ah,al
   jnz    @fin
   inc    ecx
   dec    edx
   jnz    @RCmp
   @fin:
   pop    edx
   pop    eax
end;

// Renvoie l'adresse d'un buffer de taille size rempli de 0
function GetBuffer(Size : longWord): longword; register;
asm
   push   edi
   push   eax              // rserver longueur
   call   System.@GetMem
   mov    edi,eax          // edi pointe buffer
   pop    ecx              // remise longueur
   push   eax              // sauvegarde adresse retour
   mov    edx,ecx          // calcul des itrations
   and    edx,$3           // reste div/4
   shr    ecx,2            // longueur div 4
   xor    eax,eax          // dword de 0 pour chargements
   cld                     // incrmenter edi
   rep    stosd            // boucle sur dwords complets
   mov    ecx,edx          // boucle sur octets seuls
   rep    stosb
   pop    eax              // remise adresse buffer
   pop    edi
end;

//renvoie en [eax] la chane issue d'un buffer en [esi;eax]
procedure ReString;  //en entre, eax contient la longueur du buffer.
asm
   push   edi
   push   edx
   call   SansZero          // enlve les ventuels '0' au dpart de esi
   mov    edx,eax           // compatibilit D2009
   call   System.@NewAnsiString
   mov    edi,eax
   mov    ecx,edx
   call   Recopie
   pop    edx
   pop    edi
end;

// Incrmente la valeur contenue dans un buffer point esi au dbut
// eax = len(esi) en sortie idem
procedure Plus1;
asm
   mov    ecx,eax
   @RInc:
   dec    ecx
   inc    byte ptr[esi]+ecx
   cmp    byte ptr[esi]+ecx,$64
   jb     @fin
   sub    byte ptr[esi]+ecx,$64
   jmp    @RInc
   @fin:
   test   ecx,ecx
   jge    @Retour
   neg    ecx
   sub    esi,ecx
   add    eax,ecx
   @Retour:
end;

// Dcremente la valeur en buffer esi  point dbut
// eax = len(esi) en sortie idem
procedure Moins1;
asm
   mov    ecx,eax
   @RDec:
   dec    ecx
   dec    byte ptr[esi]+ecx
   jns    @fItere
   mov    byte ptr[esi]+ecx,$63
   jmp    @RDec
   @fItere:
   test   ecx,ecx
   jz     @Retour
   call   SansZero
   @Retour:
end;

// Additionne le contenu de edi avec celui de esi points en dbut
// edi<=esi obligatoire
// eax = len(esi)  et edx = len(edi)
// Sortie rsultat en esi point dbut et eax = len(esi)
Procedure Ajoute;
asm
   push   edx
   push   eax              // conserve len(esi)
   add    edi,edx
   mov    ecx,edx          // ecx, longueur petit
   add    esi,eax          // pointeurs en fins
   mov    edx,eax          // edx, longueur grand
   xor    eax,eax
   @RAdd1:  // boucle additive des octets du petit
   mov    al,ah
   xor    ah,ah
   dec    esi
   add    al,byte ptr[esi]
   dec    edi
   add    al,byte ptr[edi]
   cmp    al,$64
   jb     @SRAdd1
   sub    al,$64
   inc    ah
   @SRAdd1:
   mov    byte ptr[esi],al
   dec    edx
   dec    ecx
   jnz    @RAdd1
   @RAdd2:  // boucle additive jusqu' fin du grand et non retenue
   mov    al,ah
   xor    ah,ah
   dec    esi
   add    al,byte ptr[esi]
   cmp    al,$64
   jb     @SRAdd2
   sub    al,$64
   inc    ah
   @SRAdd2:
   mov    byte ptr[esi],al
   dec    edx
   jg     @RAdd2
   test   ah,ah
   jnz    @RAdd2
   @fin:
   pop    eax
   test   edx,edx
   jz     @FFin
   inc    eax
   @FFin:
   pop    edx
end;

// Soustrait le contenu de edi au contenu de esi points au dbut
// edi <= esi obligatoire
// eax = len(esi) et edx = len(edi)
// Sortie rsultat en esi point dbut et eax = len(esi)
Procedure Retire;
asm
   push  edi
   push  esi
   push  edx
   push  eax
   add   edi,edx           // edi pointe sa fin + 1
   mov   ecx,edx
   add   esi,eax           // esi pointe sa fin + 1
   xor   dl,dl             // dl flag de retenue  0
   @RSub:    // boucle soustractive des octets du petit
   dec   esi
   mov   ah,byte ptr[esi]  // octet cumulateur
   dec   edi
   mov   al,byte ptr[edi]  // octet  retirer
   add   al,dl             // ajouter retenue  retirer
   xor   dl,dl             // raz retenue
   cmp   al,ah             // soustraction possible ?
   jbe   @Sous
   add   ah,$64            // non ajouter 100 au cumul
   inc   dl                // retenir 1
   @Sous:
   sub   ah,al             // soustraire
   mov   byte ptr[esi],ah
   dec   ecx
   jnz   @RSub
   @Rfin:                  // dernires retenues
   test  dl,dl
   jz    @Sortie
   xor   dl,dl
   dec   esi
   cmp   byte ptr[esi],$0
   jne   @SRFin
   add   byte ptr[esi],$64
   inc   dl
   @SRFin:
   dec   byte ptr[esi]
   jmp   @Rfin
   @Sortie:
   pop   eax              // Remise pointages
   pop   edx
   pop   esi
   pop   edi
   call  SansZero
end;

// multiplie edi par ebx et accumule le rsultat en esi points au dbut
// eax = len (edi) edx = len(ebx) ecx = len(esi)
// Sortie esi avec eax=len(esi); edx = len(edi) et ecx=len(ebx);
procedure Produit;
var  Adr, lbx, ldi, lsi, Cent, Dcr,Ret  : longword;
asm
   mov    Adr,esi
   mov    ldi,eax
   mov    lsi,ecx
   mov    lbx,edx
   mov    Cent,$64
   add    esi,lsi             // remise pointeur esi fin suivante
   add    ebx,edx             // ebx pointe fin + 1 multiplicateur
          // boucle multiplicative sur les octets d'ebx
          @GMul:
          mov    ecx,ldi
          add    edi,ecx      // edi pointe fin + 1
          mov    Dcr,ecx      // Dcr compteur ldi
          xor    ecx,ecx
          mov    Ret,ecx      // init retenue  0
          dec    ebx
          mov    cl,byte ptr[ebx] // octet multiplicateur
          xor    eax,eax
                 // boucle multiplicative des octets d'edi
                 @LMul:
                 dec    edi
                 mov    al,byte ptr[edi]  // octet  multiplier
                 mul    ecx
                 add    eax,Ret           // ajouter retenue
                 div    Cent
                 mov    Ret,eax           // nouvelle retenue
                 dec    esi
                 mov    al,byte ptr[esi]
                 add    eax,edx           // accumuler les units du produit
                 cmp    eax,Cent          // dpasse 100 ?
                 jb     @SLMul            // non
                 sub    eax,Cent          // oui retirer 100
                 inc    Ret               // et incrmenter retenue
                 @SLMul:
                 mov    byte ptr[esi],al
                 dec    Dcr
                 jnz    @LMul
          dec    esi                      // dcalage accumulateur d'un octet
          mov    eax,Ret                  // dernire retenue-ligne
          add    byte ptr[esi],al
          add    esi,ldi                  // remise esi
          dec    lbx                      // itrations multiplicateur
          jnz    @GMul
   // Remise des valeurs et ajustement sortie
   mov    esi,adr
   mov    eax,lsi
   @RZero:                    // chercher 1er byte significatif
   cmp    eax,1
   je     @fin
   cmp    byte ptr[esi],$0
   jne    @fin
   inc    esi                 // placer esi
   dec    eax
   jmp    @RZero
   @fin:
   mov    ecx,lbx
   mov    edx,ldi
end;

// Divise par 2 le buffer point dbut par esi
// en entre eax = len(esi) idem en sortie
procedure Div2;
asm
    push  eax               // garder longueur
    push  esi               // garder adresse
    mov   ecx,eax           // dcompteur boucle
    xor   ah,ah             // registre retenue
       @RDemi:              // boucle Div 2
       mov    al,byte ptr[esi]
       add    al,ah         // ajouter retenue
       xor    ah,ah         // raz retenue
       shr    al,1          // divise byte par 2
       jnc    @SRd          // pas de reste ?
       mov    ah,$64        // si, retenir 100
       @SRd:
       mov    byte ptr[esi],al
       inc    esi           // itration
       dec    ecx
       jnz    @RDemi
    pop   esi               // revenir au dpart
    pop   eax
    call  SansZero
end;

// double le buffer point par esi en dbut
// en entre et sortie eax = len(esi)
procedure Fois2;
asm
   add    esi,eax      //esi pointe fin + 1
   mov    ecx,eax      // ecx dcompteur de boucle
   mov    edx,eax      // compteur octets
   xor    ah,ah
     @RDouble:
     dec    esi
     mov    al,byte ptr[esi]
     shl    al,1        // dcimal x 2
     add    al,ah       // ajouter retenue
     xor    ah,ah
     cmp    al,$64      // dpasse 100 ?
     jb     @srd
     sub    al,$64      // oui nouvelle retenue
     mov    ah,$1
     @srd:
     mov   byte ptr[esi],al // replace byte
     dec   ecx              // itrations
     jnz   @RDouble
   test  ah,ah        // dernire retenue ?
   jz    @fin         // non
   dec   esi          // oui
   mov   byte ptr[esi],ah
   inc   edx
   @fin:
   mov  eax,edx
end;

// Divise le buffer esi;eax par le buffer edi;edx
// en entre ecx adresse le buffer du quotient
// Retourne le reste en esi;eax et le quotient en edi;edx
procedure Divise;
var N, Pq, Pr, Pa, lq, lr, la : longword;
asm
   // init des variables
   mov      N,0
   mov      Pr,esi
   mov      lr,eax
   mov      Pa,edi
   mov      la,edx
   mov      Pq,ecx
   mov      lq,1
   @BIter:  // boucle de calcul des itrations
   mov      edi,Pr
   mov      edx,lr
   mov      esi,Pa
   mov      eax,la
   call     Compare     // while aux<=R
   jz       @SIter
   js       @SIter
   jmp      @BDiv
   @SIter:
   call     Fois2       // aux:=2*aux
   inc      N           // N:=N+1
   mov      Pa,esi      // paramtres aux
   mov      la,eax
   jmp      @BIter
   @BDiv: // boucle de division proprement dite
   cmp      N,0         // while N>0
   je       @Retour
   dec      N           // N:=N-1
   mov      esi,Pa      // aux:=aux/2
   mov      eax,la
   call     Div2
   mov      Pa,esi
   mov      la,eax
   mov      esi,Pq      // Q:=Q*2
   mov      eax,lq
   call     Fois2
   mov      Pq,esi
   mov      lq,eax
   mov      esi,Pr      // r>=aux ?
   mov      eax,lr
   mov      edi,Pa
   mov      edx,la
   call     Compare
   jz       @SDiv2
   js       @BDiv       // non on boucle
   @SDiv2:
   call     Retire      // oui : R:=R-aux
   mov      Pr,esi
   mov      lr,eax
   mov      esi,Pq      // et Q:=Q+1
   mov      eax,lq
   call     Plus1
   mov      Pq,esi
   mov      lq,eax
   jmp      @BDiv
   @Retour:             // remise des paramtres
   mov     edi,Pq
   mov     edx,lq
   mov     esi,Pr
   mov     eax,lr
end;

// multiplie un buffer esi;eax par 10
// en sortie esi;eax en place
procedure Fois10;
var ln : longword;
asm
    push   ecx
    push   edx
    mov    ln,eax
    add    esi,eax            // esi pointe fin buffer+1
    mov    ecx,eax
    xor    edx,edx
    @RShl4:  // boucle de rotation de 4 bits  gauche de tous les bytes
    dec    esi
    mov    al,byte ptr[esi]  // charger ancien octet
    aam                      // dcompactage h.ah|b.al
    mov    dh,ah             // garder le haut en haut dx
    mov    ah,al             // passer le bas de ax en haut
    mov    al,dl             // passer le bas de dx en bas de al
    mov    dl,dh             // passer le haut de dx en bas
    aad                      // recombiner nouvel octet
    mov    byte ptr[esi],al  // et replacer
    dec    ecx
    jnz    @RShl4
    test   dl,dl             // dernier octet ?
    jz     @fin              // non
    inc    ln                // oui 1 byte de plus
    dec    esi
    mov    byte ptr[esi],dl
    @fin:                    // remise des paramtres
    mov    eax,ln
    pop    edx
    pop    ecx
end;

// divise un buffer esi;eax par 10
// sortie esi;eax en place
procedure Div10;
asm
   push    eax
   mov     ecx,eax
   xor     edx,edx
   @RShr4:
   mov     al,byte ptr[esi]
   aam
   mov     dl,al
   mov     al,ah
   mov     ah,dh
   mov     dh,dl
   aad
   mov     byte ptr[esi],al
   inc     esi
   dec     ecx
   jnz     @RShr4
   pop     eax
   sub     esi,eax
end;

// ROUTINES DE CONVERSIONS *****************************************************

// Convertit une string ASCII dcimale en GCent
function StrToGCent(St : string): GCent;
asm
   push     esi
   mov      esi,eax       // esi pointe St
   push     edi
   push     ebx
   push     edx
   mov      eax,edx
   call     System.@LStrClr
   mov      eax,[esi]-$4  // eax = len(esi)
   mov      ebx,eax       // garder en ebx
   shr      eax,1
   adc      eax,$0        // moiti arrondie suprieure
   mov      edx,eax       // compatibilit D2009
   call     System.@NewAnsiString
   pop      ecx
   mov      [ecx],eax     // Result
   mov      edi,eax
   add      edi,edx       // edi pointe fin Result +1
   add      esi,ebx       // esi pointe fin St+1
   xor      eax,eax
   @RConv:
   mov      ah,$30
   dec      esi
   dec      ebx
   mov      al,byte ptr[esi]
   test     ebx,ebx
   jz       @SConv
   dec      esi
   dec      ebx
   mov      ah,byte ptr[esi]
   @SConv:
   sub      ax,$3030
   aad
   dec      edi
   mov      byte ptr[edi],al
   dec      edx
   jnz      @RConv
   pop      ebx
   pop      edi
   pop      esi
end;

// Convertit un GCent en Chane d'ASCII dcimaux
function GCentToStr(Nb : GCent): string;
var buf,lnb : longword;
asm
   push    esi
   push    edi
   mov     edi,eax     // edi pointe Nb
   push    edx         // Result
   mov     eax,edx
   call    System.@LStrClr
   mov     eax,[edi]-4
   shl     eax,1       // buffer double
   mov     lnb,eax
   call    GetBuffer
   mov     buf,eax
   mov     esi,eax     // esi pointe buffer
   mov     edx,lnb
   mov     ecx,edx
   shr     ecx,1
   xor     eax,eax
   @RConv:            // Boucle conversion en ASCII
   mov     al,byte ptr[edi]
   inc     edi
   aam
   add     ax,$3030
   mov     byte ptr[esi],ah
   inc     esi
   mov     byte ptr[esi],al
   inc     esi
   dec     ecx
   jnz     @RConv
   mov     esi,buf
   mov     eax,edx
   @Rzed:          // liminer les zros devant
   cmp     eax,1
   je      @fin
   cmp     byte ptr[esi],$30
   jne     @fin
   inc     esi
   dec     eax
   jmp     @RZed
   @fin:
   call    ReString
   pop     edx
   mov     [edx],eax
   mov     eax,buf
   mov     edx,lnb
   call    System.@FreeMem
   pop     edi
   pop     esi
end;

// Convertit un nombre 32 bits non sign en GCent
function IntToGCent(Nb: longword): GCent; register;
asm
    push   esi
    push   edi
    push   ebx
    push   edx
    mov    ebx,eax
    mov    eax,edx
    call   System.@LStrClr
    // Initialisation du Buffer
    mov    edi,esp      // pointe un buffer de passage sur la pile
    sub    edi,$100
    // conversion
    mov    ecx,$64       // diviseur par 100
    mov    eax,ebx       // Nb  convertir
    xor    ebx,ebx       // compteur octets convertis
    @RDiv:
    xor    edx,edx
    div    ecx           // divise par cent
    dec    edi
    mov    byte ptr[edi],dl // place le reste en buffer
    inc    ebx
    cmp    eax,$64       // termin ?
    jae    @RDiv         // non
    test   eax,eax       // un reste supplmentaire ?
    jz     @fin          // non
    dec    edi           // oui
    mov    byte ptr[edi],al
    inc    ebx
    // former Result
    @fin:
    mov    eax,ebx
    call   System.@NewAnsiString
    pop    edx
    mov    [edx],eax    // Placer Adr Result
    mov    esi,edi
    mov    edi,eax
    mov    ecx,ebx
    rep    movsb
    // retour
    pop    ebx
    pop    edi
    pop    esi
end;

 // ROUTINES DE GENERATIONS RAPIDES *********************************************

// Renvoie un GCent reprsentant une valeur dcimale de NbDigits chiffres
// Si Odd=true , le GCent est obligatoirement impair sinon parit alatoire
// A utiliser aprs un Randomize;
function  RandomGCent(NbDigits : longword; const Odd : boolean): GCent; register;
var lnb: longword; P: byte;
asm
   push   ebx
   mov    lnb,eax              // Rserve la longueur
   mov    P,dl                 // sauve parit
   push   ecx
   mov    eax,ecx
   call   System.@LStrClr      // rinitialiser Result
   // calcul de la longueur du futur GCent
   mov    eax,lnb
   shr    eax,1                // NbDigits/2
   adc    eax,0                // arrondi unit suprieure si reste 1
   jnz    @Str
   inc    eax                  // il faut au moins un octet
   // demander la chane
   @Str:
   mov    edx,eax              // compatibilit D2009
   call   System.@NewAnsiString
   mov    ecx,edx
   pop    edx
   mov    [edx],eax           // affecter son adresse  Result
   mov    byte ptr[eax],0     // init au moins un octet 0
   mov    ebx,eax             // ebx pointeur sur Result
   add    ebx,ecx             // pointeur fin Nb+1
   mov    ecx,lnb             // compteur digit
   test   ecx,ecx
   jz     @fin                // cas o NbDigits=0
   // constituer les octets du GCent
   @RBytes:
   cmp    ecx,1               // Denier octet  un seul digit ?
   je     @DerOctet
   mov    eax,$64             // non alors fourniture alatoire [0..99]
   {$IFDEF VER200}
   call    System.Random
   {$ELSE}
   call    System.@RandInt
   {$ENDIF}
   dec    ebx
   cmp    ecx,lnb             // 1er octet de la fin ?
   jne    @SBytes
   or     al,P                // oui, ajuster la parit
   @SBytes:
   mov    byte ptr[ebx],al
   sub    ecx,2
   jz     @fin
   jmp    @RBytes
   @DerOctet:
   mov    eax,9               // un seul digit  [1..9]
   {$IFDEF VER200}
   call    System.Random
   {$ELSE}
   call    System.@RandInt
   {$ENDIF}
   inc    eax
   dec    ebx
   mov    byte ptr[ebx],al
   @fin:
   pop    ebx
end;

// ROUTINES DE TESTS ***********************************************************

// Compare deux strings GCent
// retour Negatif si N1<N2 ; 0 si N1 = N2 ; positif si N1 > N2
function CompGCent(Nv1,Nv2 : string): shortInt; register;
asm
   push  esi
   mov   esi,eax
   push  edi
   mov   edi,edx
   mov   eax,[esi]-$4
   mov   edx,[edi]-$4
   call  compare
   mov   al,-1
   js    @fin
   mov   al,1
   jnz   @fin
   mov   al,0
   @fin:
   pop   edi
   pop   esi
end;

// Teste la parit d'un GCent
// renvoie true si Pair, false si impair
function GCentPair(Nv: GCent): boolean; register;
asm
   mov   ecx,[eax]-4
   mov   al,byte ptr[eax]+ecx-1
   bt    eax,0
   setnc al
end;

// CALCULS SUR GCent ***********************************************************

// Renvoie un GCent lev de 10^N
function GCentMulN10(Nv : GCent; N : longword): GCent; register;
var Buf,lb,ln,lad,Nn : longword;
asm
   push  esi
   mov   esi,eax      // esi pointe Nv
   push  edi
   push  ecx          // garder Result
   mov   Nn,edx       // garder multiplicateur
   mov   eax,[esi]-4
   mov   ln,eax       // garder longueur Nv
   shr   edx,1        // calculer Nb byte 0
   mov   lad,edx
   add   eax,lad
   inc   eax          // Plus 1 pour ventuel impair
   mov   lb,eax
   call  GetBuffer
   mov   Buf,eax
   mov   edi,Buf
   inc   edi
   mov   ecx,ln
   call  Recopie
   mov   esi,Buf
   inc   esi
   bt    Nn,0         // exposant impair ?
   jnc   @fin         // non
   mov   eax,ln       // oui donc fois dix
   call  Fois10
   @fin:
   mov   esi,Buf
   mov   eax,lb
   call  Restring
   mov   esi,eax
   pop   eax
   call  System.@LStrClr
   mov   [eax],esi    // Result
   pop   edi
   pop   esi
   mov   eax,Buf      // librer mmoire
   mov   edx,lb
   call  System.@FreeMem
end;

// Renvoie un GCent divis de 10^N
function GCentDivN10(Nv : GCent; N : longword): GCent; register;
var Buf,lb,ln,Nn : longword;
asm
   push    esi
   mov     esi,eax
   mov     Nn,edx
   push    edi
   push    ecx             // Clear Result
   mov     eax,[esi]-4
   mov     ln,eax
   mov     ecx,eax
   shl     ecx,1           // nombre de chiffres suffisants ?
   cmp     edx,ecx
   jl      @PlusGrand      // oui
   @RetZero:               // non retour #0
   mov     eax,1
   mov     lb,eax
   call    GetBuffer
   mov     Buf,eax
   mov     esi,Buf
   mov     eax,1
   jmp     @fin
   @PlusGrand:            // oui, retour tronqu de N chiffres
   shr     edx,1
   sub     eax,edx
   mov     lb,eax
   call    GetBuffer
   mov     Buf,eax
   mov     edi,eax
   mov     ecx,lb
   call    Recopie
   mov     esi,Buf
   mov     eax,lb
   bt      Nn,0
   jnc     @fin
   call    Div10
   @fin:
   call    Restring
   mov     esi,eax
   pop     eax
   call    System.@LStrClr
   mov     [eax],esi    // Result
   pop     edi
   pop     esi
   mov     eax,Buf      // librer mmoire
   mov     edx,lb
   call    System.@FreeMem
end;

// Renvoie le nombre de 0 de la fin du GCent
function NbZeroGCent(Nv: GCent): longword; register;
asm
   push  esi
   mov   esi,eax
   mov   ecx,[eax]-4
   add   esi,ecx
   xor   edx,edx      // compteur 0
   @Rbytes0:
   dec   esi
   cmp   byte ptr[esi],0
   jnz   @DerZero
   add   edx,$2
   dec   ecx
   jz    @fin
   jmp   @Rbytes0
   @DerZero: // test du byte non nul pour voir si ses units=0
   mov   al,byte ptr[esi]
   aam
   test  al,al
   jnz   @fin
   inc   edx
   @fin:
   pop   esi
   mov   eax,edx
end;

// Renvoie une GCent Nv augmente de 1
function Incremente(Nv: GCent): GCent; register;
var buffer, lnb : longword;
asm
   push   edi
   push   esi
   mov    esi,eax         //esi pointe Nv
   push   edx             // RAZ Result
   mov    eax,[esi]-4
   mov    lnb,eax
   inc    eax             // buffer de len(Nv) + 1
   call   GetBuffer
   mov    buffer,eax
   mov    edi,eax         // edi pointe buffer
   inc    edi
   mov    ecx,lnb
   call   Recopie
   mov    esi,buffer
   inc    esi
   mov    eax,lnb
   call   Plus1
   call   Restring
   mov    esi,eax
   pop    eax
   call   System.@LStrClr
   mov    [eax],esi      // Result
   pop    esi
   pop    edi
   mov    eax,buffer
   mov    edx,lnb
   inc    edx
   call   System.@FreeMem
end;

// Renvoie un GCent Nv Diminue de 1
function Decremente(Nv: GCent): GCent; register;
var buffer, lnb : longword;
asm
   push   edi
   push   esi
   mov    esi,eax          // esi pointe Nv
   push   edx              // Raz Result
   mov    eax,[esi]-4
   mov    lnb,eax          // buffer de len(Nv)
   call   GetBuffer
   mov    buffer,eax
   mov    edi,eax          // edi pointe buffer
   mov    ecx,lnb
   call   Recopie
   mov    esi,buffer
   mov    eax,lnb
   call   Moins1
   call   Restring
   mov    esi,eax
   pop    eax
   call   System.@LStrClr
   mov   [eax],esi
   pop    esi
   pop    edi
   mov    eax,buffer
   mov    edx,lnb
   call   System.@FreeMem
end;

// Renvoie un GCent moiti de l'entre
function Demi(Nv : GCent): GCent; register;
var buffer,lnb: longword;
asm
    push    esi
    mov     esi,eax      // edi pointe Nv
    push    edi
    push    edx
    mov     eax,[esi]-$4 // demander un buffer de len(Nv)
    mov     lnb,eax
    call    GetBuffer
    mov     edi,eax      // edi pointe buffer
    mov     buffer,eax
    mov     ecx,lnb
    call    Recopie
    mov     esi,buffer
    mov     eax,lnb
    call    Div2
    call    Restring
    mov     esi,eax
    pop     eax
    call    System.@LStrClr
    mov    [eax],esi   // Result
    mov     eax,buffer
    mov     edx,lnb
    call    System.@FreeMem
    pop     edi
    pop     esi
end;

// Renvoie un GCent double de l'entre
function Double(Nv: GCent): GCent; register;
var buffer, lnb : longword;
asm
    push    esi
    mov     esi,eax      // esi pointe Nv
    push    edi
    push    edx
    mov     eax,[esi]-$4 // demander un buffer de len(Nv)+1
    inc     eax
    mov     lnb,eax
    call    GetBuffer
    mov     edi,eax
    mov     buffer,eax
    mov     ecx,lnb
    dec     ecx
    inc     edi          // edi pointe buffer+1
    call    Recopie
    mov     esi,buffer
    inc     esi
    mov     eax,lnb
    dec     eax
    call    Fois2
    call    Restring
    mov     esi,eax
    pop     eax
    call    System.@LStrClr
    mov    [eax],esi   // Result
    mov     eax,buffer
    mov     edx,lnb
    call    System.@FreeMem
    pop     edi
    pop     esi
end;

// Renvoie la Chane de NV1 + NV2 avec NV1>=NV2 obligatoire
function Additionne(Nv1,Nv2: GCent): GCent; register;
var buffer, lnb: longword;
asm
   push   esi
   mov    esi,eax        // esi pointe Nv1
   push   edi
   mov    edi,edx
   push   ecx            // Rserve adresse pour Result
   push   edi            // Rserve adresse Nv2
   mov    eax,[esi]-4
   inc    eax            // buffer de len(Nv1)+1
   mov    lnb,eax
   call   GetBuffer
   mov    buffer,eax
   mov    edi,eax
   inc    edi            // edi pointe buffer + 1
   mov    ecx,lnb
   dec    ecx
   call   Recopie        // Copie Nv1 en buffer
   pop    edi            // edi pointe Nv2
   mov    edx,[edi]-4    // edx, len(edi)
   mov    esi,buffer
   mov    eax,lnb
   call   Ajoute
   call   ReString
   mov    esi,eax
   pop    eax
   call   System.@LStrClr
   mov   [eax],esi
   pop    edi
   pop    esi
   mov    edx,lnb        // Librer buffer
   mov    eax,buffer
   call   System.@FreeMem
end;

// Renvoie la chane  Nv1-Nv2 avec Nv1 >= Nv2 obligatoire
function Soustrait(Nv1,Nv2: GCent): GCent; register;
var buffer, lnb: longword;
asm
   push   esi
   mov    esi,eax        // esi pointe Nv1
   push   edi
   mov    edi,edx
   push   ecx            // Rserve adresse pour Result
   push   edi            // Rserve adresse Nv2
   mov    eax,[esi]-4
   mov    lnb,eax
   call   GetBuffer      // buffer de len(Nv1)
   mov    buffer,eax
   mov    edi,eax        // edi pointe buffer + 1
   mov    ecx,lnb
   call   Recopie        // Copie Nv1 en buffer
   pop    edi            // edi pointe Nv2
   mov    edx,[edi]-4    // edx, len(edi)
   mov    esi,buffer
   mov    eax,lnb
   call   Retire
   call   ReString
   mov    esi,eax
   pop    eax
   call   System.@LStrClr
   mov   [eax],esi       // Result
   mov    edx,lnb        // Librer buffer
   mov    eax,buffer
   call   System.@FreeMem
   pop    edi
   pop    esi
end;

// Renvoie le produit de deux chane ASCII  Nv1,Nv2
function  Multiplie(Nv1,Nv2 : GCent): GCent; register;
var  buffer,lnb : longword;
asm
   push   edi
   mov    edi,eax        // placements pointeurs
   push   esi
   push   ebx
   mov    ebx,edx
   push   ecx            // retenue Result
   mov    eax,[edi]-4
   add    eax,[ebx]-4
   inc    eax            // buffer de lnb1+lnb2+1
   mov    lnb,eax
   call   GetBuffer
   mov    buffer,eax
   mov    esi,eax
   mov    eax,[edi]-4
   mov    edx,[ebx]-4
   mov    ecx,lnb
   call   Produit
   call   Restring
   mov    esi,eax
   pop    eax
   call   System.@LStrClr
   mov   [eax],esi        // Result
   pop    ebx
   pop    esi
   pop    edi
   mov    eax,buffer      // librer buffer
   mov    edx,lnb
   call   System.@FreeMem
end;

// Renvoie le quotient en Q et le reste en R de Nv1/nv2
procedure DivMod(Nv1,Nv2 : GCent; var Q,R : GCent); register;
var Buf,lnb,lb,AdrDe,AdrDr,AdrQ,AdrR,Pr,lr,Paux,laux : longword;
asm
  // sauvegardes
  push   esi
  push   edi
  mov    AdrDe,eax      // adresse Dividende
  mov    AdrDr,edx      // adresse Diviseur
  mov    AdrQ,ecx       // adresse du retour quotient
  mov    eax,ecx
  call   System.@LStrClr
  mov    eax,[ebp+$8]
  mov    AdrR,eax       // adresse du retour reste
  call   System.@LStrClr
  // demander un buffer de 3 fois len(dividende)*2
  mov    edi,AdrDe
  mov    ecx,3
  mov    eax,[edi]-4
  shl    eax,1
  mov    lb,eax         // lb = capacit d'un buffer
  mul    ecx
  mov    lnb,eax        // lnb = capacit totale
  call   GetBuffer
  mov    Buf,eax
  // recopier Dividende en fin buf1
  mov    esi,AdrDe
  mov    ecx,[esi]-4
  mov    edi,Buf
  add    edi,lb
  sub    edi,ecx
  mov    Pr,edi         // Pr;lr buffer reste
  mov    lr,ecx
  call   Recopie
  // recopier Diviseur en fin de buf2
  mov    esi,AdrDr
  mov    ecx,[esi]-4
  mov    edi,Buf
  add    edi,lb
  add    edi,lb
  sub    edi,ecx
  mov    Paux,edi
  mov    laux,ecx       // Paux;laux buffer auxilliaire
  call   Recopie
  // appeler la routine de division binaire
  mov    ecx,Buf        // ecx adresse du buffer quotient
  add    ecx,lnb
  dec    ecx
  mov    esi,Pr
  mov    eax,lr
  mov    edi,Paux
  mov    edx,laux
  call   Divise
  // Renvoyer les strings
  call   Restring        // Reste
  mov    esi,AdrR
  mov    [esi],eax
  mov    esi,edi         // Quotient
  mov    eax,edx
  call   Restring
  mov    esi,AdrQ
  mov    [esi],eax
  // librer la mmoire
  mov    eax,Buf
  mov    edx,lnb
  call   System.@FreeMem
  pop    edi
  pop    esi
end;

// renvoie un GCent multipli N fois par 100 (ultra rapide);
function ShlNGCent(Nv : GCent; N : longword): Gcent; register
var lnv, lN : longword;
asm
   push    esi
   push    edi
   push    ecx           // garder Result
   mov     lN,edx        // garder multiplicateur
   mov     esi,eax       // esi pointe Nv
   mov     eax,[esi]-$04 // eax = len(Nv)
   mov     lnv,eax
   add     eax,edx       // longueur nouvelle string
   call    System.@NewAnsiString
   mov     edi,eax       // edi pointe nouvelle string
   pop     eax           // remise result
   call    System.@LStrClr
   mov     [eax],edi
   mov     ecx,Lnv       // recopier le nombre
   call    Recopie
   mov     ecx,Ln        // ajouter les 0
   xor     eax,eax
   rep     stosb
   pop     edi
   pop     esi
end;

// renvoie un GCent divis N fois par 100 (ultra rapide);
function ShrNGCent(Nv : GCent; N : longword): Gcent; register
  var lnv : longword;
  const nulle : GCent = #0;
asm
   push    esi
   push    edi
   push    ecx           // garder Result
   mov     esi,eax       // esi pointe Nv
   mov     eax,[esi]-$04 // eax = len(Nv)
   sub     eax,edx       // longueur nouvelle string
   jg      @SShr
   mov     eax,1         // si <= 0, retour 0
   mov     esi,nulle
   @SShr:
   mov     lnv,eax
   call    System.@NewAnsiString
   mov     edi,eax       // edi pointe nouvelle string
   pop     eax           // remise result
   call    System.@LStrClr
   mov     [eax],edi
   mov     ecx,Lnv       // recopier le nombre
   call    Recopie
   pop     edi
   pop     esi
end;

// Renvoie 10  la puissance Exp
function  DixPowerN(Exp : longword): GCent; register;
var Buf,LBuf : longword;
asm
  push    esi
  mov     ecx,1         // init 1er octet  1 
  shr     eax,1         // calcul Nb octets retour
  jnc     @S1       
  mov     ecx,$A        // 1er octet  dix si Exp Impair
  @S1:
  push    edx          // garde result
  push    ecx          // garde 1er octet
  inc     eax          // longueur de la chane
  mov     LBuf,eax
  call    GetBuffer    // buffer plein de 0
  mov     Buf,eax
  mov     esi,eax
  pop     eax          // Rappel 1er octet
  mov     byte ptr[esi],al
  mov     eax,LBuf
  call    Restring
  mov     esi,eax      // Chane retour en esi
  pop     eax          // Rappel Result
  call    System.@LStrClr // Raz Result
  mov     [eax],esi    // Nouvelle chane en Result
  pop     esi
  mov     eax,Buf      // librer le Buffer
  mov     edx,LBuf
  call    System.@FreeMem
end;

// AJOUTS :

function PuissanceGCent(Nb: GCent; Exp : longword): GCent; 
//       Rcursive : Renvoie Nb^Exp avec Exp >= 1
var      tmp : GCent; ModExp: byte;
begin    Result:=#1;
         if Exp=0 then Exit else
         if Exp=1 then begin Result:=Nb; Exit; end;
         asm shr   Exp,1
             setC  al
             mov   ModExp,al
         end;
         tmp:=PuissanceGCent(Nb, Exp);
         Result:=Multiplie(tmp,tmp);
         if ModExp=1 then Result:=Multiplie(Result,Nb);
end;

procedure TrimD0GCent(var Nv : GCent; var nz : int64);
begin     nz := NbZeroGCent(Nv);
          if nz>0 then Nv := GCentDivN10(Nv,nz);
end;

function AddGCent(Nv1,Nv2: GCent): GCent;
begin    if CompGCent(Nv1,Nv2) >= 0
         then Result:=Additionne(Nv1,Nv2)
         else Result:=Additionne(Nv2,Nv1);
end;

function DifGCent(Nv1,Nv2: GCent): GCent;
begin    if CompGCent(Nv1,Nv2) >= 0
         then Result:=Soustrait(Nv1,Nv2)
         else Result:=Soustrait(Nv2,Nv1);
end;

procedure DivModGCent(Nv1,Nv2 : GCent; var Q,R : GCent);
begin     if Nv2=#0
          then begin ShowMessage('Tentative de division par zro'); Q:=#0; R:=#0; Exit; end
          else DivMod(Nv1,Nv2, Q,R);
end;

// Calcule la racine carre d'un GCent par l'algotithme dit de la potence
function Racine100(A : GCent) : GCent;
   var Reste,TR,chiffre,Ch,z,y,sy : GCent; Index,lg,t: longword;
begin
   Reste:=#0; Result:=Reste;
   Index:=1;  lg:=Length(A);
   repeat
      if (Index=1) and (odd(lg)) then t:=1 else t:=2;
      TR:=Copy(A,Index,t);
      if Reste<>#0 then Reste:=Reste+TR else Reste:=TR;
      Chiffre:=#0; Ch:=#0; y:=#0; sy:=#0;
      if Reste>#0 then begin
         z:=Double(Result)+#0; // ici une astuce qui remplace la multiplication par #2#0 de z:=Multiplie(Result,DeuxFoisBase);
         while CompGCent(Reste,y)>=0 do begin
           sy:=y; Chiffre:=Ch;
           ch:=incremente(Ch);
           y:=AddGCent(z,Ch); y:=Multiplie(y,Ch);
         end;
         Reste:=Soustrait(Reste,sy);
      end;
      Result:=Result+Chiffre;
      Index:=Index+t;
   until Index>=lg;
end; // Racine100

// Calcule la racine carre entire par dfaut d'un GCent;
// Algorithme de Hron avec semence en potence
function RacineGCent(Nb : GCent): GCent;
var      Z,R : GCent; lg,n,np,m,k : longword;
begin
  if Nb=#0 then begin Result:=Nb; exit; end;
  lg:=length(Nb);
  if lg<9 then Result:=Racine100(Nb)  // J'ai pris 9 arbitrairement, mais on peut relativiser en fonction des tests
  else begin
       // Pour entte Quart :
       k:=3*(lg div 4);
       if odd(k) then np:=k-1 else np:=k-2;
       m:=lg - np;
       n:=np div 2;
       R:=Racine100(Copy(Nb,1,m));    // calcul de l'entte approche
       R:=ShlNGCent(Incremente(R),n); // ajout de la partie en 00
       nbToursGCent:=0;
       repeat                         // boucles du Hron
         Result:=R; inc(nbToursGCent);
         DivMod(Nb,R,R,Z);
         R:=Demi(AddGCent(Result,R));
       until CompGCent(R,Result)>=0;
  end; // Hron
end; // RacineGCent

function GCentToInt64(Nb: GCent): Int64;
begin    Result:=StrToInt64(GCentToStr(Nb));
end;

END. // end of unit

